from __future__ import annotations


from scaleway_core.profile import Profile
from scaleway_core.validations.string_validation import (
    is_access_key,
    is_organization_id,
    is_project_id,
    is_region,
    is_secret_key,
    is_url,
    is_zone,
)


class Client(Profile):
    _request_count: int = 1

    @staticmethod
    def from_profile(profile: Profile) -> Client:
        return Client(**profile.__dict__)

    def _increment_request_count(self) -> int:
        current = self._request_count
        self._request_count += 1
        return current

    def validate(self) -> bool:
        return self.validate_credentials() and self.validate_settings()

    def validate_credentials(self) -> bool:
        if not self.access_key or not self.secret_key:
            raise ValueError(
                "Invalid secrets, access_key & secret_key must be defined. See https://www.scaleway.com/docs/console/my-project/how-to/generate-api-key"
            )

        if not is_access_key(self.access_key):
            raise ValueError(
                f"Invalid access key format '{self.access_key}', expected SCWXXXXXXXXXXXXXXXXX format. See https://www.scaleway.com/docs/console/my-project/how-to/generate-api-key"
            )

        if not is_secret_key(self.secret_key):
            raise ValueError(
                f"Invalid secret key format '{self.secret_key}', expected a UUID: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx. See https://www.scaleway.com/docs/console/my-project/how-to/generate-api-key"
            )

        return True

    def validate_settings(self) -> bool:
        if self.default_organization_id is not None:
            if not self.default_organization_id:
                raise ValueError("Default organization ID cannot be empty")

            if not is_organization_id(self.default_organization_id):
                raise ValueError(
                    f"Invalid organization ID format '${self.default_organization_id}', expected a UUID: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
                )

        if self.default_project_id is not None:
            if not self.default_project_id:
                raise ValueError("Default project ID cannot be empty")

            if not is_project_id(self.default_project_id):
                raise ValueError(
                    f"Invalid project ID format '${self.default_project_id}', expected a UUID: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"
                )

        if self.default_region is not None:
            if not self.default_region:
                raise ValueError("Default region cannot be empty")

            if not is_region(self.default_region):
                raise ValueError(
                    f"Invalid default region format '{self.default_region}'"
                )

        if self.default_zone is not None:
            if not self.default_zone:
                raise ValueError("Default zone cannot be empty")

            if not is_zone(self.default_zone):
                raise ValueError(f"Invalid default zone format '{self.default_zone}'")

        if not is_url(self.api_url):
            raise ValueError(f"Invalid URL '{self.api_url}'")

        if self.api_url[-1] == "/":
            raise ValueError(
                f"Invalid URL ${self.api_url}: it should not have a trailing slash"
            )

        if self.default_page_size is not None:
            if self.default_page_size < 1:
                raise ValueError(
                    f"Invalid defaultPageSize {self.default_page_size}: it should be a number above 0"
                )

        return True



from __future__ import annotations

import dataclasses
import logging
import os
from dataclasses import dataclass
from typing import Optional, Type, TypeVar

import yaml
from scaleway_core import __version__
from scaleway_core.profile.file import CONFIG_PROPERTIES_TO_PROFILE

from .env import ENV_KEY_SCW_CONFIG_PATH, ENV_KEY_SCW_PROFILE, ENV_VARIABLES_TO_PROFILE


@dataclass
class ProfileDefaults:
    default_organization_id: Optional[str] = None
    """
    Your organization ID is the identifier of your account inside Scaleway infrastructure.
    """

    default_project_id: Optional[str] = None
    """
    Your project ID is the identifier of the project your resources are attached to.
    """

    default_region: Optional[str] = None
    """
    A region is represented as a geographical area such as France (Paris) or the Netherlands (Amsterdam).
    It can contain multiple availability zones.

    Examples: fr-par, nl-ams.
    """

    default_zone: Optional[str] = None
    """
    A region can be split into many availability zones (AZ).
    Latency between multiple AZ of the same region are low as they have a common network layer.

    Examples: fr-par-1, nl-ams-1
    """

    default_page_size: Optional[int] = None
    """
    The default number of results when requesting a paginated resource.
    """


@dataclass
class ProfileConfig:
    access_key: Optional[str] = None
    """
    You need an access key and a secret key to connect to Scaleway API.
    Generate your access key at the following address: https://console.scaleway.com/project/credentials.
    """

    secret_key: Optional[str] = None
    """
    The secret key is the value that can be used to authenticate against the API (the value used in X-Auth-Token HTTP-header).
    The secret key MUST remain secret and not given to anyone or published online.
    Generate your secret key at the following address: https://console.scaleway.com/project/credentials.
    """

    api_url: str = "https://api.scaleway.com"
    """
    The Scaleway API URL.
    Change that if you want to direct requests to a different endpoint.
    """

    api_allow_insecure: bool = False
    """
    Allow insecure connection to the API.
    """

    user_agent: str = f"scaleway-sdk-python/{__version__}"
    """
    The User-Agent sent with each request.
    """


ProfileSelf = TypeVar("ProfileSelf", bound="Profile")


@dataclass
class Profile(ProfileDefaults, ProfileConfig):
    def merge(self, other: Profile) -> None:
        """
        Merge the current profile with another one.
        """
        for field in dataclasses.fields(Profile):
            current_value = getattr(self, field.name)

            if current_value is None:
                setattr(self, field.name, getattr(other, field.name))

    @classmethod
    def from_env(cls: Type[ProfileSelf], force_none: bool = False) -> ProfileSelf:
        """
        Loads profile from environment variables.
        """
        profile = cls()
        for env_variable, profile_property in ENV_VARIABLES_TO_PROFILE.items():
            value = os.environ.get(env_variable)
            if value is not None:
                setattr(profile, profile_property, value)
            elif force_none:
                setattr(profile, profile_property, None)

        return profile

    @classmethod
    def get_default_config_directory(cls) -> str:
        xdg_config_path = os.environ.get("XDG_CONFIG_HOME")
        if xdg_config_path is not None and xdg_config_path != "":
            return os.path.join(xdg_config_path, "scw")

        return os.path.join(os.path.expanduser("~"), ".config", "scw")

    @classmethod
    def get_default_config_file_path(cls, filepath: Optional[str] = None) -> str:
        if filepath is not None:
            return filepath

        filepath = os.environ.get(ENV_KEY_SCW_CONFIG_PATH)
        if filepath is not None and filepath != "":
            return filepath

        return os.path.join(Profile.get_default_config_directory(), "config.yaml")

    @classmethod
    def from_config_file(
        cls: Type[ProfileSelf],
        filepath: Optional[str] = None,
        profile_name: Optional[str] = "default",
        force_none: bool = False,
    ) -> ProfileSelf:
        filepath = cls.get_default_config_file_path(filepath)

        with open(filepath, "r") as f:
            config = yaml.safe_load(f)

            if not isinstance(config, dict):
                raise ValueError("Invalid config file")

            profile = cls()
            for file_property, profile_property in CONFIG_PROPERTIES_TO_PROFILE.items():
                value = config.get(file_property)
                if value is not None:
                    setattr(profile, profile_property, value)
                elif force_none:
                    setattr(profile, profile_property, None)

            if profile_name is not None and profile_name != "default":
                has_profile = (
                    "profiles" in config
                    and isinstance(config["profiles"], dict)
                    and profile_name in config["profiles"]
                )

                if not has_profile:
                    raise ValueError(f"Profile '{profile_name}' not found")

                overrides = config["profiles"][profile_name]

                if not isinstance(overrides, dict):
                    raise ValueError(f"Invalid profile '{profile_name}'")

                for (
                    file_property,
                    profile_property,
                ) in CONFIG_PROPERTIES_TO_PROFILE.items():
                    value = overrides.get(file_property)
                    if value is not None:
                        setattr(profile, profile_property, value)
                    elif force_none:
                        setattr(profile, profile_property, None)

            return profile

    @classmethod
    def from_config_file_and_env(
        cls: Type[ProfileSelf],
        filepath: Optional[str] = None,
        profile_name: Optional[str] = os.environ.get(ENV_KEY_SCW_PROFILE, "default"),
    ) -> ProfileSelf:
        """
        Loads profile from a config file and environment variables.

        Environment variables override config file.
          - If config file is not found, the profile is still loaded from environment variables.
          - If you want it to throw an error in case of missing or invalid config file, use `Profile.from_config_file` and `Profile.from_env` instead.
        """

        has_config_profile = False
        try:
            config_profile = cls.from_config_file(filepath, profile_name)
            has_config_profile = True
        except Exception as e:
            logging.getLogger("scaleway").warning(
                f"Could not load profile from config file: {e}"
            )

        env_profile = cls.from_env(force_none=has_config_profile)
        if has_config_profile:
            env_profile.merge(config_profile)

        return env_profile



CONFIG_PROPERTIES_TO_PROFILE = {
    "access_key": "access_key",
    "secret_key": "secret_key",
    "api_url": "api_url",
    "default_organization_id": "default_organization_id",
    "default_project_id": "default_project_id",
    "default_region": "default_region",
    "default_zone": "default_zone",
}



from typing import Dict, Optional


def project_or_organization_id(
    organization_id: Optional[str],
    project_id: Optional[str],
    default_project_id: Optional[str],
) -> Dict[str, str]:
    if project_id is not None:
        return {
            "project_id": project_id,
        }

    if organization_id is not None:
        return {
            "organization_id": organization_id,
        }

    if default_project_id is not None:
        return {
            "project_id": default_project_id,
        }

    raise ValueError("You must provide either a project_id or an organization_id")



Zone = str
ZONE_FR_PAR_1 = Zone("fr-par-1")
ZONE_FR_PAR_2 = Zone("fr-par-2")
ZONE_FR_PAR_3 = Zone("fr-par-3")
ZONE_NL_AMS_1 = Zone("nl-ams-1")
ZONE_NL_AMS_2 = Zone("nl-ams-2")
ZONE_NL_AMS_3 = Zone("nl-ams-3")
ZONE_PL_WAW_1 = Zone("pl-waw-1")
ZONE_PL_WAW_2 = Zone("pl-waw-2")

ALL_ZONES = [
    ZONE_FR_PAR_1,
    ZONE_FR_PAR_2,
    ZONE_FR_PAR_3,
    ZONE_NL_AMS_1,
    ZONE_NL_AMS_2,
    ZONE_NL_AMS_3,
    ZONE_PL_WAW_1,
    ZONE_PL_WAW_2,
]



from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class ServiceInfo:
    """
    Represents API metadata.

    These metadata are only here for debugging. Do not rely on these values.
    """

    name: str
    """
    Name of the API
    """

    description: str
    """
    Human readable description for the API.
    """

    version: str
    """
    Version of the API.
    """

    documentation_url: Optional[str] = None
    """
    Web url where the documentation of the API can be found.
    """


def unmarshal_ServiceInfo(data: Any) -> ServiceInfo:
    """
    Unmarshals a ServiceInfo object from a dict.
    """
    if not isinstance(data, dict):
        raise TypeError(
            "Unmarshalling the type 'ServiceInfo' failed as data isn't a dictionary."
        )

    return ServiceInfo(
        name=data["name"],
        description=data["description"],
        version=data["version"],
        documentation_url=data.get("documentation_url", None),
    )


def marshal_ServiceInfo(obj: ServiceInfo) -> Dict[str, Any]:
    """
    Marshals a ServiceInfo object into a dict.
    """
    return {
        "name": obj.name,
        "description": obj.description,
        "version": obj.version,
        "documentation_url": obj.documentation_url,
    }



from .fetch_all_pages import (
    fetch_all_pages,
    fetch_all_pages_async,
)

from .project_or_organization_id import project_or_organization_id

from .random_name import random_name

from .resolve_one_of import OneOfPossibility, resolve_one_of

from .strenummeta import StrEnumMeta

from .validate_path_param import validate_path_param

from .waiter import (
    WaitForOptions,
    WaitForStopCondition,
    wait_for_resource,
    wait_for_resource_async,
)

__all__ = [
    "fetch_all_pages",
    "fetch_all_pages_async",
    "project_or_organization_id",
    "random_name",
    "OneOfPossibility",
    "resolve_one_of",
    "StrEnumMeta",
    "validate_path_param",
    "WaitForOptions",
    "WaitForStopCondition",
    "wait_for_resource",
    "wait_for_resource_async",
]



from enum import EnumMeta
from typing import Any, Optional


class StrEnumMeta(EnumMeta):
    def __call__(
        cls,
        value: str,
        names: Optional[Any] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if names is not None:
            return super().__call__(value, names, *args, **kwargs)

        try:
            # attempt to get an enum member
            return super().__call__(value, names, *args, **kwargs)
        except ValueError:
            # no such member exists, but we don't care if the value is a string
            if not isinstance(value, str):
                raise ValueError(f"{value} is not a valid {cls.__name__} or string")

        return value



import asyncio
import inspect
import math
import random
import time
from dataclasses import dataclass
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    Optional,
    TypeVar,
    Union,
)

system_random = random.SystemRandom()

T = TypeVar("T")
U = TypeVar("U")

WaitForStopCondition = Callable[[T], U]


@dataclass
class WaitForOptions(Generic[T, U]):
    """
    The options to wait until a resource is ready.
    """

    timeout: float = 300
    """
    Timeout in seconds.

    :default: 300 seconds (5 minutes).
    """

    min_delay: float = 1
    """
    The minimum delay before the next try in seconds.

    :default: 1 second.
    """

    max_delay: float = 30
    """
    The maximum delay before the next try in seconds.

    :default: 30 seconds.
    """

    stop: Optional[WaitForStopCondition[T, U]] = None
    """
    The condition to stop trying.

    :default: Waits for non-transient value.
    """


def _exponential_backoff_strategy(
    min_delay: float,
    max_delay: float,
) -> Generator[float, None, None]:
    """
    Generates a sequence of delays using an exponential backoff strategy.

    :param min_delay: The minimum delay before the next try in seconds.
    :param max_delay: The maximum delay before the next try in seconds.
    """
    if min_delay < 1:
        raise ValueError("min_delay must be greater than 1")

    if max_delay < min_delay:
        raise ValueError("max_delay must be greater than min_delay")

    max = math.log(max_delay / min_delay) / math.log(2) + 1

    attempt = 1
    while True:
        if attempt > max:
            yield max_delay
        else:
            yield system_random.uniform(min_delay, min_delay * 2 ** (attempt - 1))

        attempt += 1


def _delayed_loop(options: WaitForOptions[Any, Any]) -> Generator[float, None, None]:
    strategy = _exponential_backoff_strategy(options.min_delay, options.max_delay)
    timeout_time = time.time() + options.timeout

    while time.time() <= timeout_time:
        delay = next(strategy)
        if timeout_time <= (time.time() + delay):
            break

        yield delay

    raise TimeoutError("Timeout while waiting for resource")


async def wait_for_resource_async(
    fetcher: Callable[..., Awaitable[T]],
    options: WaitForOptions[T, Union[bool, Awaitable[bool]]],
    args: Dict[str, Any],
) -> T:
    """
    Fetches resource several times until an expected condition is reached, timeouts, or throws an exception.
    """
    if options.stop is None:
        raise ValueError("options.stop is required")

    for delay in _delayed_loop(options):
        await asyncio.sleep(delay)

        resource = await fetcher(**args)

        should_stop = options.stop(resource)
        if inspect.isawaitable(should_stop):
            should_stop = await should_stop

        if should_stop:
            return resource

    raise TimeoutError("Timeout while waiting for resource")


def wait_for_resource(
    fetcher: Callable[..., T],
    options: WaitForOptions[T, bool],
    args: Dict[str, Any],
) -> T:
    """
    Fetches resource several times until an expected condition is reached, timeouts, or throws an exception.
    """
    if options.stop is None:
        raise ValueError("options.stop is required")

    for delay in _delayed_loop(options):
        time.sleep(delay)

        resource = fetcher(**args)

        if options.stop(resource):
            return resource

    raise TimeoutError("Timeout while waiting for resource")



from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, TypeVar

T = TypeVar("T")


@dataclass
class OneOfPossibility(Generic[T]):
    param: str

    value: Optional[T]

    default: Optional[T] = None


def resolve_one_of(
    possibilities: List[OneOfPossibility[Any]], is_required: bool = False
) -> Dict[str, Any]:
    """
    Resolves the ideal parameter and value amongst an optional list.
    """

    # Get the first non-empty parameter
    for possibility in possibilities:
        if possibility.value is not None:
            return {possibility.param: possibility.value}

    # Get the first non-empty default
    for possibility in possibilities:
        if possibility.default is not None:
            return {possibility.param: possibility.default}

    # If required, raise an error
    if is_required:
        possibilities_keys = " or ".join(
            [possibility.param for possibility in possibilities]
        )
        raise ValueError(f"one of ${possibilities_keys} must be present")

    # Else, return an empty dict
    return {}



from typing import Any, Awaitable, Callable, Dict, List, Optional, Type, TypeVar

T = TypeVar("T")


def _build_fetcher_args(
    args: Dict[str, Any],
    page: Optional[int],
) -> Dict[str, Any]:
    """
    Builds the arguments to pass to the fetcher function.
    """
    return {
        **args,
        "page": page or args.get("page") or 1,
    }


def fetch_all_pages(
    type: Type[T],
    key: str,
    fetcher: Callable[..., T],
    args: Dict[str, Any],
    page: Optional[int] = None,
) -> List[Any]:
    """
    :param key: The key to use to get the list of items from the response
    :param fetcher: The function to call to fetch the response
    :return: The list of items
    """
    fetcher_args = _build_fetcher_args(args, page)
    page = fetcher_args.get("page") or 1
    page_size = fetcher_args.get("page_size")

    data = fetcher(**fetcher_args)
    if not data:
        return []

    items: List[Any] = getattr(data, key)

    if page_size is not None and len(items) < page_size:
        return items

    if not items:
        return items

    return items + fetch_all_pages(
        type=type,
        key=key,
        fetcher=fetcher,
        args=args,
        page=page + 1,
    )


async def fetch_all_pages_async(
    type: Type[T],
    key: str,
    fetcher: Callable[..., Awaitable[T]],
    args: Dict[str, Any],
    page: Optional[int] = None,
) -> List[Any]:
    """
    :param key: The key to use to get the list of items from the response
    :param fetcher: The function to call to fetch the response
    :return: The list of items
    """
    fetcher_args = _build_fetcher_args(args, page)
    page = fetcher_args.get("page") or 1
    page_size = fetcher_args.get("page_size")

    data = await fetcher(**fetcher_args)
    if not data:
        return []

    items: List[Any] = getattr(data, key)

    if page_size is not None and len(items) < page_size:
        return items

    if not items:
        return items

    return items + await fetch_all_pages_async(
        type=type,
        key=key,
        fetcher=fetcher,
        args=args,
        page=page + 1,
    )



from .money import Money
from .money import unmarshal_Money
from .money import marshal_Money

from .region import Region
from .region import ALL_REGIONS
from .zone import Zone
from .zone import ALL_ZONES

from .scwfile import ScwFile
from .scwfile import unmarshal_ScwFile
from .scwfile import marshal_ScwFile

from .serviceinfo import ServiceInfo
from .serviceinfo import unmarshal_ServiceInfo
from .serviceinfo import marshal_ServiceInfo

from .timeseries import TimeSeriesPoint
from .timeseries import TimeSeries
from .timeseries import unmarshal_TimeSeries
from .timeseries import marshal_TimeSeries
from .timeseries import unmarshal_TimeSeriesPoint
from .timeseries import marshal_TimeSeriesPoint

__all__ = [
    "Money",
    "unmarshal_Money",
    "marshal_Money",
    "Region",
    "ALL_REGIONS",
    "Zone",
    "ALL_ZONES",
    "ScwFile",
    "unmarshal_ScwFile",
    "marshal_ScwFile",
    "ServiceInfo",
    "unmarshal_ServiceInfo",
    "marshal_ServiceInfo",
    "TimeSeriesPoint",
    "TimeSeries",
    "unmarshal_TimeSeries",
    "marshal_TimeSeries",
    "unmarshal_TimeSeriesPoint",
    "marshal_TimeSeriesPoint",
]



from datetime import datetime
from typing import Optional, Union


def validate_path_param(name: str, value: Optional[Union[str, int, datetime]]) -> str:
    """
    Returns the parameter if it's valid as path parameter
    (string and not empty, or number), else throws an exception.
    """
    if type(value) is datetime:
        value = value.isoformat()

    if (value is None) or (isinstance(value, str) and not value):
        raise ValueError(f"Path parameter {name} is required")

    return str(value)



from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class Money:
    """
    Represents an amount of money with its currency type.
    """

    currency_code: str
    """
    3-letter currency code defined in ISO 4217.
    """

    units: float
    """
    Whole units of the amount.

    For example if `currency_code` is `"USD"`, then 1 unit is one US dollar.
    """

    nanos: int
    """
    Number of nano (10^-9) units of the amount.

    The value must be between -999,999,999 and +999,999,999 inclusive.
    If `units` is positive, `nanos` must be positive or zero.
    If `units` is zero, `nanos` can be positive, zero, or negative.
    If `units` is negative, `nanos` must be negative or zero.
    For example $-1.75 is represented as `units`=-1 and `nanos`=-750,000,000.
    """


def unmarshal_Money(data: Any) -> Money:
    """
    Unmarshal an instance of Money from the given data.
    """
    if not isinstance(data, dict):
        raise TypeError(
            "Unmarshalling the type 'Money' failed as data isn't a dictionary."
        )

    return Money(
        currency_code=data["currency_code"],
        units=data["units"],
        nanos=data["nanos"],
    )


def marshal_Money(data: Money) -> Dict[str, Any]:
    """
    Marshal an instance of Money into a JSON compatible data structure.
    """
    return {
        "currency_code": data.currency_code,
        "units": data.units,
        "nanos": data.nanos,
    }



import json
import logging
from dataclasses import dataclass
from typing import IO, Any, Dict, Iterable, List, Mapping, Optional, Tuple, Union

import requests
from requests import Response

from .client import Client

Body = Union[
    str, bytes, Mapping[str, Any], Iterable[Tuple[str, Optional[str]]], IO[Any]
]

Params = Mapping[str, Any]


@dataclass
class APILogger:
    logger: logging.Logger

    identifier: int

    def log_request(
        self,
        method: str,
        url: str,
        params: List[Tuple[str, Any]],
        headers: Dict[str, str],
        body: Optional[str],
    ) -> None:
        if not self.logger.isEnabledFor(logging.DEBUG):
            return

        raw_params = "&".join([f"{k}={v}" for (k, v) in params])

        debug_headers = {
            **headers,
            "x-auth-token": "*****",
        }

        title = (
            f"--------------- Scaleway SDK REQUEST {self.identifier} ---------------\n"
        )

        log = f"{title}\n"
        log += f"{method} {url}?{raw_params}\n"

        for k, v in debug_headers.items():
            log += f"{k}: {v}\n"

        if body is not None:
            log += "\n"
            log += body
            log += "\n"

        log += "-" * len(title)

        self.logger.debug(log)

    def log_response(
        self,
        response: requests.Response,
    ) -> None:
        if not self.logger.isEnabledFor(logging.DEBUG):
            return

        title = (
            f"--------------- Scaleway SDK RESPONSE {self.identifier} ---------------"
        )
        ok = "OK" if response.ok else "NOK"

        log = f"{title}\n"
        log += f"HTTP {response.status_code} {ok}\n"

        for k, v in response.headers.items():
            log += f"{k}: {v}\n"

        if response.text:
            log += f"\n{response.text}\n"

        log += "-" * len(title)

        self.logger.debug(log)


@dataclass
class ScalewayException(Exception):
    response: Response

    @property
    def status_code(self) -> int:
        return self.response.status_code

    def __str__(self) -> str:
        return self.response.text


@dataclass
class ValidationError(ScalewayException):
    errors: Dict[str, str]


class API:
    def __init__(self, client: Client, *, bypass_validation: bool = False):
        if not bypass_validation:
            client.validate()

        self.client = client
        self._log = logging.getLogger("scaleway")

    def _request(
        self,
        method: str,
        path: str,
        params: Params = {},
        headers: Dict[str, str] = {},
        body: Optional[Body] = None,
    ) -> requests.Response:
        additional_headers: Dict[str, str] = {}

        method = method.upper()
        if method == "POST" or method == "PUT" or method == "PATCH":
            additional_headers["Content-Type"] = "application/json; charset=utf-8"

            if body is None:
                body = {}

        raw_body = json.dumps(body) if body is not None else None

        request_params: List[Tuple[str, Any]] = []
        for k, v in params.items():
            if v is None:
                continue

            if isinstance(v, list):
                for item in v:
                    request_params.append((k, item))
            else:
                request_params.append((k, v))

        headers = {
            "accept": "application/json",
            "x-auth-token": self.client.secret_key or "",
            "user-agent": self.client.user_agent,
            **additional_headers,
            **headers,
        }

        url = f"{self.client.api_url}{path}"

        logger = APILogger(self._log, self.client._increment_request_count())

        logger.log_request(
            method=method,
            url=url,
            params=request_params,
            headers=headers,
            body=raw_body,
        )

        response = requests.request(
            method=method,
            url=url,
            params=request_params,
            headers=headers,
            data=raw_body,
            verify=not self.client.api_allow_insecure,
        )

        logger.log_response(
            response=response,
        )

        return response

    def _throw_on_error(self, res: requests.Response) -> None:
        if res.status_code >= 400:
            data = res.json()

            if data:
                if "message" in data:
                    message = data["message"]

                    if message == "Validation Error":
                        raise ValidationError(res, data["fields"])

                    raise ScalewayException(res)

                raise ScalewayException(res)

            raise ScalewayException(res)



from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class ScwFile:
    """
    Represents a Scaleway file.
    """

    name: str
    """
    Name of the file.
    """

    content_type: str
    """
    Content-type of the file.
    """

    content: str
    """
    Content of the file in base64.
    """


def unmarshal_ScwFile(data: Any) -> ScwFile:
    """
    Unmarshals a ScwFile object from a dict.
    """
    if not isinstance(data, dict):
        raise TypeError(
            "Unmarshalling the type 'ScwFile' failed as data isn't a dictionary."
        )

    return ScwFile(
        name=data["name"],
        content_type=data["content_type"],
        content=data["content"],
    )


def marshal_ScwFile(obj: ScwFile) -> Dict[str, Any]:
    """
    Marshals a ScwFile object into a dict.
    """
    return {
        "name": obj.name,
        "content_type": obj.content_type,
        "content": obj.content,
    }



ENV_KEY_SCW_CONFIG_PATH = "SCW_CONFIG_PATH"
ENV_KEY_SCW_PROFILE = "SCW_PROFILE"
ENV_KEY_SCW_ACCESS_KEY = "SCW_ACCESS_KEY"
ENV_KEY_SCW_SECRET_KEY = "SCW_SECRET_KEY"  # nosec B105
ENV_KEY_SCW_API_URL = "SCW_API_URL"
ENV_KEY_SCW_DEFAULT_ORGANIZATION_ID = "SCW_DEFAULT_ORGANIZATION_ID"
ENV_KEY_SCW_DEFAULT_PROJECT_ID = "SCW_DEFAULT_PROJECT_ID"
ENV_KEY_SCW_DEFAULT_REGION = "SCW_DEFAULT_REGION"
ENV_KEY_SCW_DEFAULT_ZONE = "SCW_DEFAULT_ZONE"

ENV_VARIABLES_TO_PROFILE = {
    ENV_KEY_SCW_ACCESS_KEY: "access_key",
    ENV_KEY_SCW_SECRET_KEY: "secret_key",
    ENV_KEY_SCW_API_URL: "api_url",
    ENV_KEY_SCW_DEFAULT_ORGANIZATION_ID: "default_organization_id",
    ENV_KEY_SCW_DEFAULT_PROJECT_ID: "default_project_id",
    ENV_KEY_SCW_DEFAULT_REGION: "default_region",
    ENV_KEY_SCW_DEFAULT_ZONE: "default_zone",
}



import re
import uuid
from urllib.parse import urlparse

_is_access_key_regex = re.compile(r"^SCW[A-Z0-9]{17}$", re.IGNORECASE)


def is_access_key(s: str) -> bool:
    return _is_access_key_regex.match(s) is not None


def _is_uuid(s: str) -> bool:
    try:
        uuid.UUID(s)
        return True
    except ValueError:
        return False


def is_secret_key(s: str) -> bool:
    return _is_uuid(s)


def is_organization_id(s: str) -> bool:
    return _is_uuid(s)


def is_project_id(s: str) -> bool:
    return _is_uuid(s)


_is_region_regex = re.compile(r"^[a-z]{2}-[a-z]{3}$", re.IGNORECASE)


def is_region(s: str) -> bool:
    return _is_region_regex.match(s) is not None


_is_zone_regex = re.compile(r"^[a-z]{2}-[a-z]{3}-[1-9]$", re.IGNORECASE)


def is_zone(s: str) -> bool:
    return _is_zone_regex.match(s) is not None


def is_url(s: str) -> bool:
    try:
        result = urlparse(s)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False



from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List
from dateutil import parser


@dataclass
class TimeSeriesPoint:
    """
    Represents a point in a TimeSeries.
    """

    timestamp: datetime
    """
    Date of the point.
    """

    value: float
    """
    Value of the point.
    """


def unmarshal_TimeSeriesPoint(data: Any) -> TimeSeriesPoint:
    """
    Unmarshal an instance of TimeSeriesPoint from the given data.
    """
    if not isinstance(data, dict):
        raise TypeError(
            "Unmarshalling the type 'TimeSeriesPoint' failed as data isn't a dictionary."
        )

    return TimeSeriesPoint(
        timestamp=parser.isoparse(data["timestamp"]),
        value=data["value"],
    )


def marshal_TimeSeriesPoint(data: TimeSeriesPoint) -> Dict[str, Any]:
    """
    Marshal an instance of TimeSeriesPoint into a JSON compatible data structure.
    """
    return {
        "timestamp": data.timestamp.isoformat(),
        "value": data.value,
    }


@dataclass
class TimeSeries:
    """
    Represents a time series that could be used for graph purposes.
    """

    name: str
    """
    Name of the metric.
    """

    points: List[TimeSeriesPoint]
    """
    Points contains all the points that composed the series.
    """

    metadata: Dict[str, str]
    """
    Metadata contains some string metadata related to a metric.
    """


def unmarshal_TimeSeries(data: Any) -> TimeSeries:
    """
    Unmarshal an instance of TimeSeries from the given data.
    """
    if not isinstance(data, dict):
        raise TypeError(
            "Unmarshalling the type 'TimeSeries' failed as data isn't a dictionary."
        )

    return TimeSeries(
        name=data["name"],
        points=[unmarshal_TimeSeriesPoint(point) for point in data["points"]],
        metadata=data["metadata"],
    )


def marshal_TimeSeries(data: TimeSeries) -> Dict[str, Any]:
    """
    Marshal an instance of TimeSeries into a JSON compatible data structure.
    """
    return {
        "name": data.name,
        "points": [marshal_TimeSeriesPoint(point) for point in data.points],
        "metadata": data.metadata,
    }



from .env import (
    ENV_KEY_SCW_ACCESS_KEY,
    ENV_KEY_SCW_API_URL,
    ENV_KEY_SCW_CONFIG_PATH,
    ENV_KEY_SCW_DEFAULT_ORGANIZATION_ID,
    ENV_KEY_SCW_DEFAULT_PROJECT_ID,
    ENV_KEY_SCW_DEFAULT_REGION,
    ENV_KEY_SCW_DEFAULT_ZONE,
    ENV_KEY_SCW_SECRET_KEY,
)
from .profile import Profile, ProfileConfig, ProfileDefaults

__all__ = [
    "Profile",
    "ProfileConfig",
    "ProfileDefaults",
    "ENV_KEY_SCW_CONFIG_PATH",
    "ENV_KEY_SCW_ACCESS_KEY",
    "ENV_KEY_SCW_SECRET_KEY",
    "ENV_KEY_SCW_API_URL",
    "ENV_KEY_SCW_DEFAULT_ORGANIZATION_ID",
    "ENV_KEY_SCW_DEFAULT_PROJECT_ID",
    "ENV_KEY_SCW_DEFAULT_REGION",
    "ENV_KEY_SCW_DEFAULT_ZONE",
]
